/* cpu_mem_ras.c COPYRIGHT FUJITSU LIMITED 2018-2021 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/irqdomain.h>
#include <linux/interrupt.h>
#include <linux/printk.h>
#ifdef __DEBUG
#pragma GCC push_options
#pragma GCC optimize ("O2")
#include <linux/irq.h>
#pragma GCC pop_options
#endif
#include <linux/acpi.h>
#include <linux/fdtable.h>
#include <linux/kprobes.h>
#include <linux/sched/signal.h>
#include <linux/version.h>
#include <linux/kdebug.h>
#include <asm/esr.h>
#include <asm/sysreg.h>
#include <asm/traps.h>

/*
 * debug mode
 */
#ifdef __DEBUG
#define panic(msg)	printk(msg)
#endif

/*
 * module param
 */
struct policy_info {
	char **policy_str;
	int array_size;
	int *policy_flag;
};

/* node policy */
#define NODE_POLICY_STR_INDEX_CONT	0
#define NODE_POLICY_STR_INDEX_STOP	1
static const char *node_policy_str[] = {
	[NODE_POLICY_STR_INDEX_CONT] = "continuation",
	[NODE_POLICY_STR_INDEX_STOP] = "stop",
};

static int node_policy_set(const char *str, const struct kernel_param *kp);
const struct kernel_param_ops param_ops_node_policy = {
	.set = node_policy_set,
	.get = param_get_charp,
	.free = param_free_charp,
};

static const char *node_policy = NULL;
static int node_policy_flag;
module_param_cb(node_policy, &param_ops_node_policy, &node_policy, 0600);

/* uae store policy */
#define UAE_STORE_POLICY_STR_INDEX_IGNORE	0
#define UAE_STORE_POLICY_STR_INDEX_LOG		1
#define UAE_STORE_POLICY_STR_INDEX_DUMP		2
static const char *uae_store_policy_str[] = {
	[UAE_STORE_POLICY_STR_INDEX_IGNORE] = "ignore",
	[UAE_STORE_POLICY_STR_INDEX_LOG] = "log",
	[UAE_STORE_POLICY_STR_INDEX_DUMP] = "dump",
};

static int uae_store_policy_set(const char *str, const struct kernel_param *kp);
const struct kernel_param_ops param_ops_uae_store_policy = {
	.set = uae_store_policy_set,
	.get = param_get_charp,
	.free = param_free_charp,
};

static const char *uae_store_policy = NULL;
static int uae_store_policy_flag;
module_param_cb(uae_store_policy, &param_ops_uae_store_policy, &uae_store_policy, 0600);

/* uae store error log interval */
static int uae_store_log_interval_set(const char *str, const struct kernel_param *kp);
const struct kernel_param_ops param_ops_uae_store_log_interval = {
	.set = uae_store_log_interval_set,
	.get = param_get_long,
};
#define UAE_STORE_LOG_INTERVAL_DEFAULT	1
static long uae_store_log_interval = UAE_STORE_LOG_INTERVAL_DEFAULT;
static bool uae_store_log_flag = 0;
spinlock_t uae_store_log_lock;
module_param_cb(uae_store_log_interval, &param_ops_uae_store_log_interval,
		&uae_store_log_interval, 0600);
#define UAE_STORE_LOG_INTERVAL_MAX	600
#define UAE_STORE_LOG_INTERVAL_MIN	0

/* uae store count */
static int uae_store_count_get(char* str, const struct kernel_param *kp);
const struct kernel_param_ops param_ops_uae_store_count = {
	.get = uae_store_count_get,
};
atomic_t uae_store_count_atomic = ATOMIC_INIT(0);
static unsigned long uae_store_count = 0;
module_param_cb(uae_store_count, &param_ops_uae_store_count, &uae_store_count, 0400);

/*
 * debug option
 */
#define DO_SERROR_DETECT	(0x01)
#define DO_SEA_DETECT		(0x02)
#define HSWE_DETECT		(0x04)
#define GSWE_DETECT		(0x08)
static unsigned long error_detect_enable = 0x7;
module_param(error_detect_enable, ulong, 0600);

static bool hswe_panic_enable = false;
module_param(hswe_panic_enable, bool, 0600);

/*
 * work queue
 */
static struct workqueue_struct *uae_store_log_wq;
static void uae_store_log_handler(struct work_struct *work);
static DECLARE_DELAYED_WORK(uae_store_log_work, uae_store_log_handler);

/* panic flag */
static bool panic_flag = 0;

/* cpus */
#define FX1000_CPUS		60
#define GENERIC_A64FX_CPUS	48

/*
 * logging
 */

#define ras_debug_msg(msgid, fmt, ...)	\
	pr_debug("[DEBUG] CPU-MEM-RAS: %04d: " fmt, msgid, ##__VA_ARGS__)
#define ras_info_msg(msgid, fmt, ...)	\
	pr_info("[INFO] CPU-MEM-RAS: %04d: " fmt, msgid, ##__VA_ARGS__)
#define ras_err_msg(msgid, fmt, ...)	\
	pr_err("[ERROR] CPU-MEM-RAS: %04d: " fmt, msgid, ##__VA_ARGS__)
#define ras_notice_msg(msgid, fmt, ...)	\
	pr_notice("[NOTE] CPU-MEM-RAS: %04d: " fmt, msgid, ##__VA_ARGS__)
#define ras_warn_msg(msgid, fmt, ...)	\
	pr_warn("[WARN] CPU-MEM-RAS: %04d: " fmt, msgid, ##__VA_ARGS__)
#define ras_emerg_msg(msgid, fmt, ...)	\
	pr_emerg("[EMERG] CPU-MEM-RAS: %04d: " fmt, msgid, ##__VA_ARGS__)

/* debug message */
#define DEBUG_MSG_GET_HOST_VIRQ		0
#define DEBUG_MSG_GET_GUEST_VIRQ	1
#define DEBUG_MSG_GUEST_SW_ERR_COMM	2
#define DEBUG_MSG_GUEST_SW_ERR_FDFILE	3

/* info message */
#define INFO_MSG_ERRSTATUS_NO_NEED_CLEAR	0

/* error message */
#define ERR_MSG_HOST_SW_ERR_VIRQ_FAILED				0
#define ERR_MSG_HOST_SW_ERR_REQUEST_IRQ_FAILED			1
#define ERR_MSG_GUEST_SW_ERR_VIRQ_FAILED			2
#define ERR_MSG_GUEST_SW_ERR_REQUEST_IRQ_FAILED			3
#define ERR_MSG_ACPI_REGISTER_GSI_FAILED			4
#define ERR_MSG_IOREMAP_FAILD_GICv3_GICD			5
#define ERR_MSG_HOST_SW_ERR_IOREMAP_FAILD_GICv3_ITS_CNTL_REG	6
#define ERR_MSG_REGISTER_RAS_HANDLER				7
#define ERR_MSG_ERROR_PENDING_ITS				8
#define ERR_MSG_ERROR_PENDING_ICC				9
#define ERR_MSG_ERROR_PENDING_ICH_ICV				10
#define ERR_MSG_ERRSTATUS_IERR_UE_RAW_L1			11
#define ERR_MSG_ERRSTATUS_IERR_UE_RAW_L2			12
#define ERR_MSG_UE_CLEAR_FAILED					13
#define ERR_MSG_VALID_CLEAR_FAILED				14
#define ERR_MSG_ASSERTION_FAILURE				15
#define ERR_MSG_INTERNAL_DATA_PATH				16
#define ERR_MSG_INTERNAL_CONTROL_REG				17
#define ERR_MSG_DG_L1_TLB					18
#define ERR_MSG_DG_L2						19
#define ERR_MSG_UNCORRECTABLE_ADDR_ERR_STORE			20
#define ERR_MSG_GPR_SP_REG_ERR					21
#define ERR_MSG_FP_SIMD_VECTOR_PRE_REG_ERR			22
#define ERR_MSG_ERR_CODE_TYPE_INVALID_SERROR			23
#define ERR_MSG_MARKED_UE					24
#define ERR_MSG_UNCORRECTABLE_ADDR_ERR_LOAD			25
#define ERR_MSG_HOST_SW_ERR_PENDING_NONE			26
#define ERR_MSG_HOST_SW_ERR_DETECTED				27
#define ERR_MSG_ERR_CODE_TYPE_INVALID_MEMORY_ABOORT		28
#define ERR_MSG_FAILED_TO_DETECT_ERROR				29
#define ERR_MSG_REGISTER_KPROBE_FAILED				30
#define ERR_MSG_CREATE_QUEUE_FAILED				31
#define ERR_MSG_UNSUPPORTED_ENVIRONMENT				32
#define ERR_MSG_MISSING_MODULE_SYMBOL				33

#define RAS_HOOK_FUNC_SERROR	"do_serror"
#define RAS_HOOK_FUNC_SEA	"do_sea"

/*
 * function
 */
/* check node policy */
static const char *check_node_policy(const char *str, struct policy_info *pi)
{
	const char *cp = NULL;
	int len, i;

	if (pi == NULL) {
		return cp;
	}

	cp = memchr(str, '\n', strlen(str));
	if (cp) {
		len = cp - str;
	}
	else {
		len = strlen(str);
	}

	cp = NULL;
	for (i = 0; i < pi->array_size; i++) {
		if (pi->policy_str[i] == NULL) {
			continue;
		}

		if (len == strlen(pi->policy_str[i]) &&
			strncmp(str, pi->policy_str[i], len) == 0) {
				cp = pi->policy_str[i];
				*pi->policy_flag = i;
				break;
		}
	}
	return cp;
}

/* set node policy */
static int node_policy_set(const char *str, const struct kernel_param *kp)
{
	if (likely(node_policy != NULL)) {
		struct policy_info node_policy_info;
		const char *cp = NULL;

		node_policy_info.policy_str = (char **)node_policy_str;
		node_policy_info.array_size = ARRAY_SIZE(node_policy_str);
		node_policy_info.policy_flag = &node_policy_flag;
		cp = check_node_policy(str, &node_policy_info);
		if (cp) {
			param_set_charp(cp, kp);
		}
		else {
			return -EINVAL;
		}
	}
	else {
		param_set_charp(str, kp);
	}
	return 0;
}

/* set uae store policy */
static int uae_store_policy_set(const char *str, const struct kernel_param *kp)
{
	if (likely(uae_store_policy != NULL)) {
		struct policy_info uae_store_policy_info;
		const char *cp = NULL;

		uae_store_policy_info.policy_str = (char **)uae_store_policy_str;
		uae_store_policy_info.array_size = ARRAY_SIZE(uae_store_policy_str);
		uae_store_policy_info.policy_flag = &uae_store_policy_flag;
		cp = check_node_policy(str, &uae_store_policy_info);
		if (cp) {
			param_set_charp(cp, kp);
		}
		else {
			return 0;
		}
	}
	else {
		param_set_charp(str, kp);
	}

	return 0;
}

/* set uae store log interval */
static int uae_store_log_interval_set(const char *str, const struct kernel_param *kp)
{
	int ret;
	long interval;

	ret = kstrtol(str, 10, &interval);
	if (ret) {
		return 0;
	}

	if (interval < UAE_STORE_LOG_INTERVAL_MIN || interval > UAE_STORE_LOG_INTERVAL_MAX) {
		return 0;
	}

	param_set_long(str, kp);
	return 0;
}

/* get uae store count */
static int uae_store_count_get(char* str, const struct kernel_param *kp)
{
	long count = 0;

	count = (long)atomic_read(&uae_store_count_atomic);
	snprintf(str, PAGE_SIZE, "%ld\n", count);

	return strlen(str);
}

static void uae_store_log_handler(struct work_struct *work)
{
	spin_lock(&uae_store_log_lock);
	uae_store_log_flag = 0;
	spin_unlock(&uae_store_log_lock);
}


/* register bit defines */
/* ESR.ISS.AET */
#define ESR_ELx_ISS_AET_UEU		(0x01) /* Unrecoverable error */
#define ESR_ELx_ISS_AET_UEO		(0x02) /* Restartable error */
#define ESR_ELx_ISS_AET_UER		(0x03) /* Recoverable error */
#define ESR_ELx_ISS_AET_SHIFT		(10)
#define ESR_ELx_ISS_AET_MASK		(UL(0x7) << ESR_ELx_ISS_AET_SHIFT)
#define ESR_ELx_ISS_AET(esr)		((esr & ESR_ELx_ISS_AET_MASK) >> ESR_ELx_ISS_AET_SHIFT)

#define ESR_ELx_ISS_SET_UER		(0x00) /* Recoverable error */
#define ESR_ELx_ISS_SET_SHIFT		(11)
#define ESR_ELx_ISS_SET_MASK		(UL(0x3) << ESR_ELx_ISS_SET_SHIFT)
#define ESR_ELx_ISS_SET(esr)		((esr & ESR_ELx_ISS_SET_MASK) >> ESR_ELx_ISS_SET_SHIFT)

#define SYS_ERRSELR_EL1			sys_reg(3, 0, 5, 3, 1)
#define SYS_ERXSTATUS_EL1		sys_reg(3, 0, 5, 4, 2)
#define SYS_ERXADDR_EL1			sys_reg(3, 0, 5, 4, 3)

/* ERRSTATUS */
#define ERRSTATUS_SERR_MASK		(UL(0xff))
#define ERRSTATUS_SERR(estatus)		(estatus & ERRSTATUS_SERR_MASK)
#define ERRSTATUS_SERR_IMPL_ERR		(0x01)
#define ERRSTATUS_SERR_ASSR_FAIL	(0x04)
#define ERRSTATUS_SERR_INT_DATA_PATH	(0x05)
#define ERRSTATUS_SERR_INT_CNTL_REG	(0x11)

#define ERRSTATUS_IERR_UE_RAW_L1	(0x01)
#define ERRSTATUS_IERR_UE_RAW_L2	(0x02)
#define ERRSTATUS_IERR_DG_L1TLB		(0x04)
#define ERRSTATUS_IERR_DG_L2		(0x05)
#define ERRSTATUS_IERR_MARKED_UE	(0x07)
#define ERRSTATUS_IERR_UE_ADDR_ERR	(0x08)
#define ERRSTATUS_IERR_GPR_SP_ERR	(0x10)
#define ERRSTATUS_IERR_FP_SIMD_ERR	(0x11)
#define ERRSTATUS_IERR_SHIFT		(8)
#define ERRSTATUS_IERR_MASK		(UL(0x1f) << ERRSTATUS_IERR_SHIFT)
#define ERRSTATUS_IERR(estatus)		((estatus & ERRSTATUS_IERR_MASK) >> ERRSTATUS_IERR_SHIFT)

#define ERRSTATUS_MV_SHIFT		(26)
#define ERRSTATUS_MV_MASK		(UL(0x1) << ERRSTATUS_MV_SHIFT)

#define ERRSTATUS_OF_SHIFT		(27)
#define ERRSTATUS_OF_MASK		(UL(0x1) << ERRSTATUS_OF_SHIFT)

#define ERRSTATUS_UE_SHIFT		(29)
#define ERRSTATUS_UE_MASK		(UL(0x1) << ERRSTATUS_UE_SHIFT)

#define ERRSTATUS_V_SHIFT		(30)
#define ERRSTATUS_V_MASK		(UL(0x1) << ERRSTATUS_V_SHIFT)

#define ERRSTATUS_AV_SHIFT		(31)
#define ERRSTATUS_AV_MASK		(UL(0x1) << ERRSTATUS_AV_SHIFT)

#define ERRSTATUS_VALID_BITS		(ERRSTATUS_MV_MASK | ERRSTATUS_V_MASK | ERRSTATUS_AV_MASK)

/* ERRADDR */
#define ERRADDR_PADDR_MASK		(UL(0xffffffffffffff))

/* Tofu register address region */
struct tofu_reg_addr {
	unsigned long start;
	unsigned long end;
};

const struct tofu_reg_addr tofu_addr[] = {
	[0] = { /* Control Queue TNI0CQ[00-11] */
		.start	= 0x0040000000,
		.end	= 0x00400BFFFF,
	},
	[1] = { /* Barrier Channel TNI0BCH[00-15] */
		.start	= 0x0040E00000,
		.end	= 0x0040EFFFFF,
	},
	[2] = { /* Control Queue TNI1CQ[00-11] */
		.start	= 0x0041000000,
		.end	= 0x00410BFFFF,
	},
	[3] = { /* Barrier Channel TNI1BCH[00-15] */
		.start	= 0x0041E00000,
		.end	= 0x0041EFFFFF,
	},
	[4] = { /* Control Queue TNI2CQ[00-11] */
		.start	= 0x0042000000,
		.end	= 0x00420BFFFF,
	},
	[5] = { /* Barrier Channel TNI2BCH[00-15] */
		.start	= 0x0042E00000,
		.end	= 0x0042EFFFFF,
	},
	[6] = { /* Control Queue TNI3CQ[00-11] */
		.start	= 0x0043000000,
		.end	= 0x00430BFFFF,
	},
	[7] = { /* Barrier Channel TNI3BCH[00-15] */
		.start	= 0x0043E00000,
		.end	= 0x0043EFFFFF,
	},
	[8] = { /* Control Queue TNI4CQ[00-11] */
		.start	= 0x0044000000,
		.end	= 0x00440BFFFF,
	},
	[9] = { /* Barrier Channel TNI4BCH[00-15] */
		.start	= 0x0044E00000,
		.end	= 0x0044EFFFFF,
	},
	[10] = { /* Control Queue TNI5CQ[00-11] */
		.start	= 0x0045000000,
		.end	= 0x00450BFFFF,
	},
	[11] = { /* Barrier Channel TNI5BCH[00-15] */
		.start	= 0x0045E00000,
		.end	= 0x0045EFFFFF,
	},
};

/* read and clear ERR0STATUS register */
static u64 read_and_clear_err_0_status_register(void)
{
	u64 value = 0;
	u64 check_value = 0;
	const unsigned int number = 0;

	/* setting read register number and read ERR0STATUS */
	write_sysreg_s(number, SYS_ERRSELR_EL1);
	value = read_sysreg_s(SYS_ERXSTATUS_EL1);

	/* valid bit (ERR0STATUS.{MV, V, AV}) check */
	if (!(value & ERRSTATUS_VALID_BITS)) {
		/* no need to clear */
		ras_info_msg(INFO_MSG_ERRSTATUS_NO_NEED_CLEAR,
			     "no need to clear ERR0STATUS register.\n");
		goto out;
	}

	/* setting write register number clear ERR0STATUS.UE */
	write_sysreg_s(number, SYS_ERRSELR_EL1);
	write_sysreg_s(ERRSTATUS_UE_MASK, SYS_ERXSTATUS_EL1);

	/* setting read register number and read ERR0STATUS */
	write_sysreg_s(number, SYS_ERRSELR_EL1);
	check_value = read_sysreg_s(SYS_ERXSTATUS_EL1);

	/* check cleared ERR0STATUS.UE */
	if (check_value & ERRSTATUS_UE_MASK) {
		/* Because the ERR0STATUS.OF is set,
		 * clearing the ERR0STATUS.UE failed.
		 */
		ras_err_msg(ERR_MSG_UE_CLEAR_FAILED,
			    "Clearing ERR0STATUS.UE failed."
			    "ERR0STATUS=0x%llx -> 0x%llx\n", value, check_value);
		panic_flag = 1;
		goto out;
	}
	else {
		u64 temp_value;

		temp_value = check_value;
		check_value = 0;

		/* setting write register number clear ERR0STATUS.{V, AV, MV} */
		write_sysreg_s(number, SYS_ERRSELR_EL1);
		write_sysreg_s(ERRSTATUS_VALID_BITS, SYS_ERXSTATUS_EL1);
		/* setting read register number and read ERR0STATUS */
		write_sysreg_s(number, SYS_ERRSELR_EL1);
		check_value = read_sysreg_s(SYS_ERXSTATUS_EL1);

		/* check cleared ERR0STATUS.{V, AV, MV} */
		if (check_value & ERRSTATUS_VALID_BITS) {
			/* Because the ERR0STATUS.UE is set,
			 * clearing the ERR0STATUS.{V, AV, MV} failed.
			 * Because the ERR0STATUS.UE was detected
			 * before clearing the valid bit, the node is down.
			 */
			ras_err_msg(ERR_MSG_VALID_CLEAR_FAILED,
				    "Clearing valid bit failed."
				    "ERR0STATUS=0x%llx -> 0x%llx -> 0x%llx\n",
				    value, temp_value, check_value);
			panic_flag = 1;
			goto out;
		}
	}

out :
	return value;
}

static void uae_store_error_log_func(unsigned int esr, u64 err_status)
{
	bool local_flag;
	long local_interval;

	spin_lock(&uae_store_log_lock);
	local_flag = uae_store_log_flag;
	local_interval = uae_store_log_interval;

	if ((local_flag == 0) && (local_interval > 0)) {
		uae_store_log_flag = 1;
	}
	spin_unlock(&uae_store_log_lock);

	if (local_flag == 0) {
		ras_err_msg(ERR_MSG_UNCORRECTABLE_ADDR_ERR_STORE,
			    "Uncorrectable Address Error (Store) detected.\n"
			    "esr=0x%x ERR0STATUS=0x%llx count=%ld\n",
			    esr, err_status, (long)atomic_read(&uae_store_count_atomic));
		if (local_interval > 0) {
			queue_delayed_work(uae_store_log_wq, &uae_store_log_work,
					   msecs_to_jiffies(local_interval * MSEC_PER_SEC));
		}
	}
}

/* do_serror handler */
static int ras_do_serror_pre(struct kprobe *p, struct pt_regs *regs)
{
	int ret = 0;
	u64 err_status = 0;
	unsigned int esr = (unsigned int)(regs->regs[1]);

	/* EC check */
	if (ESR_ELx_EC(esr) != ESR_ELx_EC_SERROR) {
		goto out;
	}

	/* read and clear ERR0STATUS */
	err_status = read_and_clear_err_0_status_register();

	/* check error detect */
	if (!(error_detect_enable & DO_SERROR_DETECT)) {
		/* do nod detect error */
		goto out;
	}

	/* check ERR0STATUS.V */
	if (!(err_status & ERRSTATUS_V_MASK)) {
		goto out;
	}

	/* RAS Extensions Defined Error */
	if ((ESR_ELx_ISS_AET(esr) == ESR_ELx_ISS_AET_UEO) ||
	    (ESR_ELx_ISS_AET(esr) == ESR_ELx_ISS_AET_UEU) ||
	    (ESR_ELx_ISS_AET(esr) == ESR_ELx_ISS_AET_UER)) {
		switch (ERRSTATUS_SERR(err_status)) {
		case ERRSTATUS_SERR_ASSR_FAIL:
			/* no get dump */
			/* node down */
			ras_err_msg(ERR_MSG_ASSERTION_FAILURE,
				    "Assertion failure detected.\n"
				    "esr=0x%x ERR0STATUS=0x%llx\n", esr, err_status);
			panic_flag = 1;

			goto out;
			break;
		case ERRSTATUS_SERR_INT_DATA_PATH:
			/* no get dump */
			/* node down */
			ras_err_msg(ERR_MSG_INTERNAL_DATA_PATH,
				    "Internal data path detected.\n"
				    "esr=0x%x ERR0STATUS=0x%llx\n", esr, err_status);
			panic_flag = 1;

			goto out;
			break;
		case ERRSTATUS_SERR_INT_CNTL_REG:
			/* no get dump */
			/* node down */
			ras_err_msg(ERR_MSG_INTERNAL_CONTROL_REG,
				    "Internal control register detected.\n"
				    "esr=0x%x ERR0STATUS=0x%llx\n", esr, err_status);
			panic_flag = 1;

			goto out;
			break;
		default:
			break;
		}
	}

	/* ISS.AET == UEO */
	if (ESR_ELx_ISS_AET(esr) == ESR_ELx_ISS_AET_UEO) {
		/* IMPLEMENTATION DEFINED error */
		if (ERRSTATUS_SERR(err_status) == ERRSTATUS_SERR_IMPL_ERR) {
			switch (ERRSTATUS_IERR(err_status)) {
			case ERRSTATUS_IERR_UE_RAW_L1:
				/* output syslog */
				ras_err_msg(ERR_MSG_ERRSTATUS_IERR_UE_RAW_L1,
					    "Uncorrectable Error RAW L1 detected.\n");
				goto out;
				break;
			case ERRSTATUS_IERR_UE_RAW_L2:
				/* output syslog */
				ras_err_msg(ERR_MSG_ERRSTATUS_IERR_UE_RAW_L2,
					    "Uncorrectable Error RAW L2 detected.\n");
				goto out;
				break;
			case ERRSTATUS_IERR_DG_L1TLB:
				/* NODE Continuation check */
				if (node_policy_flag == NODE_POLICY_STR_INDEX_STOP) {
					/* no get dump */
					/* node down */
					ras_err_msg(ERR_MSG_DG_L1_TLB,
						    "DG L1 TLB detected.\n"
						    "esr=0x%x ERR0STATUS=0x%llx\n", esr, err_status);
					panic_flag = 1;
				}
				goto out;
				break;
			case ERRSTATUS_IERR_DG_L2:
				/* NODE Continuation check */
				if (node_policy_flag == NODE_POLICY_STR_INDEX_STOP) {
					/* no get dump */
					/* node down */
					ras_err_msg(ERR_MSG_DG_L2,
						    "DG L2 detected.\n"
						    "esr=0x%x ERR0STATUS=0x%llx\n", esr, err_status);
					panic_flag = 1;
				}
				goto out;
				break;
			case ERRSTATUS_IERR_UE_ADDR_ERR:
				atomic_inc(&uae_store_count_atomic);
				if (uae_store_policy_flag == UAE_STORE_POLICY_STR_INDEX_LOG) {
					uae_store_error_log_func(esr, err_status);
				}
				else if (uae_store_policy_flag == UAE_STORE_POLICY_STR_INDEX_DUMP) {
					/* get dump */
					/* node down */
					ras_err_msg(ERR_MSG_UNCORRECTABLE_ADDR_ERR_STORE,
						    "Uncorrectable Address Error (Store) detected.\n"
						    "esr=0x%x ERR0STATUS=0x%llx count=%ld\n",
						    esr, err_status,
						    (long)atomic_read(&uae_store_count_atomic));
					panic_flag = 1;
				}
				goto out;
				break;
			default:
				goto err_invalid;
				break;
			}
		}
		else {
			goto err_invalid;
		}
	}
	/* ISS.AET == UEU */
	else if (ESR_ELx_ISS_AET(esr) == ESR_ELx_ISS_AET_UEU) {
		/* IMPLEMENTATION DEFINED error */
		if (ERRSTATUS_SERR(err_status) == ERRSTATUS_SERR_IMPL_ERR) {
			switch (ERRSTATUS_IERR(err_status)) {
			case ERRSTATUS_IERR_GPR_SP_ERR:
				/* no get dump */
				/* node down */
				ras_err_msg(ERR_MSG_GPR_SP_REG_ERR,
					    "GPR, SP register error detected.\n"
					    "esr=0x%x ERR0STATUS=0x%llx\n", esr, err_status);
				panic_flag = 1;

				goto out;
				break;
			case ERRSTATUS_IERR_FP_SIMD_ERR:
				/* no get dump */
				/* node down */
				ras_err_msg(ERR_MSG_FP_SIMD_VECTOR_PRE_REG_ERR,
					    "FP&SIMD, vector, predicate register error detected.\n"
					    "esr=0x%x ERR0STATUS=0x%llx\n", esr, err_status);
				panic_flag = 1;

				goto out;
				break;
			default:
				goto err_invalid;
				break;
			}
		}
		else {
			goto err_invalid;
		}
	}

out:
	if (panic_flag == 1) {
		/* set esr to panic in Host-Linux processing. */
		regs->regs[1] &= ~ESR_ELx_AET;
		regs->regs[1] |= ESR_ELx_AET_UEU;
	}
	return ret;

err_invalid:
	ras_err_msg(ERR_MSG_ERR_CODE_TYPE_INVALID_SERROR,
		    "Invalid error code or type detected when SError occurred.\n"
		    "esr=0x%x ERR0STATUS=0x%llx\n", esr, err_status);
	/* set esr to panic in Host-Linux processing. */
	regs->regs[1] &= ~ESR_ELx_AET;
	regs->regs[1] |= ESR_ELx_AET_UEU;

	return ret;
}
static int ras_do_serror_fault(struct kprobe *p, struct pt_regs *regs, int trapnr)
{
	ras_err_msg(ERR_MSG_FAILED_TO_DETECT_ERROR,
		    "Failed to detect the error. [%s]\n", 
		    RAS_HOOK_FUNC_SERROR);
	return 0;
}

struct pt_regs regs_tmp[FX1000_CPUS] = {};

static void ras_check_err_addr(struct pt_regs *regs)
{
	int i, array_num, cpu;
	u64 err_addr = 0;
	struct pt_regs *regs_p;
	const unsigned int number = 0;

	/* read Error Record 0 Address Register */
	write_sysreg_s(number, SYS_ERRSELR_EL1);
	err_addr = read_sysreg_s(SYS_ERXADDR_EL1);
	err_addr &= ERRADDR_PADDR_MASK;

	/* Determine if tofu region */
	array_num = sizeof(tofu_addr) / sizeof(struct tofu_reg_addr);
	for (i = 0; i < array_num; i++) {
		if ((err_addr >= tofu_addr[i].start) &&
		    (err_addr <= tofu_addr[i].end)) {
			/*
			 * For pc of kernel page fault type errors,
			 * pc is not updated, so pc is updated.
			 */
			regs_p = (struct pt_regs *)regs->regs[2];
			regs_p->pc += 0x04;

			/*
			 * Prepare pt_regs with psate as user.
			 * Replace it with the addreas
			 * of the pt_regs formal parameter.
			 */
			cpu = smp_processor_id();
			memcpy(&regs_tmp[cpu],
			       (struct pt_regs *)regs->regs[2],
			       sizeof(struct pt_regs));
			regs_tmp[cpu].pstate &= ~PSR_MODE_MASK;
			regs_tmp[cpu].pstate |= PSR_MODE_EL0t;
			regs->regs[2] = (unsigned long)&regs_tmp[cpu];
			break;
		}
	}
}

/* do_sea handler */
static int ras_do_sea_pre(struct kprobe *p, struct pt_regs *regs)
{
	int ret = 0;
	u64 err_status = 0;
	unsigned int esr = (unsigned int)(regs->regs[1]);

	/* EC check */
	if ((ESR_ELx_EC(esr) != ESR_ELx_EC_IABT_LOW) &&
	    (ESR_ELx_EC(esr) != ESR_ELx_EC_IABT_CUR) &&
	    (ESR_ELx_EC(esr) != ESR_ELx_EC_DABT_LOW) &&
	    (ESR_ELx_EC(esr) != ESR_ELx_EC_DABT_CUR)) {
		goto out;
	}

	/* read and clear ERR0STATUS */
	err_status = read_and_clear_err_0_status_register();

	/* check error detect */
	if (!(error_detect_enable & DO_SEA_DETECT)) {
		/* do not detect error */
		goto out;
	}

	/* check ERR0STATUS.V */
	if (!(err_status & ERRSTATUS_V_MASK)) {
		goto out;
	}

	/* ISS.AET == UER */
	if (ESR_ELx_ISS_SET(esr) == ESR_ELx_ISS_SET_UER) {
		/* IMPLEMENTATION DEFINED error */
		if (ERRSTATUS_SERR(err_status) == ERRSTATUS_SERR_IMPL_ERR) {
			switch (ERRSTATUS_IERR(err_status)) {
			case ERRSTATUS_IERR_MARKED_UE:
				/* no get dump */
				/* node down */
				ras_err_msg(ERR_MSG_MARKED_UE,
					    "MARKED_UE detected.\n"
					    "esr=0x%x ERR0STATUS=0x%llx\n", esr, err_status);
				goto out;
				break;
			case ERRSTATUS_IERR_UE_ADDR_ERR:
				/* get dump */
				/* node down */
				ras_err_msg(ERR_MSG_UNCORRECTABLE_ADDR_ERR_LOAD,
					    "Uncorrectable Address Error"
					    "(Load, Instruction access) detected.\n"
					    "esr=0x%x ERR0STATUS=0x%llx\n", esr, err_status);

				/* if tofu region, don't kernel panic */
				ras_check_err_addr(regs);

				goto out;
				break;
			default:
				goto err_invalid;
				break;
			}
		}
		else {
			goto err_invalid;
		}
	}

out:
	return ret;

err_invalid:
	ras_err_msg(ERR_MSG_ERR_CODE_TYPE_INVALID_MEMORY_ABOORT,
		    "Invalid error code or type detected when memory abort occurred.\n"
		    "esr=0x%x ERR0STATUS=0x%llx\n", esr, err_status);
	return ret;
}
static int ras_do_sea_fault(struct kprobe *p, struct pt_regs *regs, int trapnr)
{
	ras_err_msg(ERR_MSG_FAILED_TO_DETECT_ERROR,
		    "Failed to detect the error. [%s]\n",
		    RAS_HOOK_FUNC_SEA);
	return 0;
}

/* kprobe */
static struct kprobe kp_do_serror = {
	.symbol_name	= RAS_HOOK_FUNC_SERROR,
	.pre_handler	= ras_do_serror_pre,
	.fault_handler	= ras_do_serror_fault,
};

static struct kprobe kp_do_sea = {
	.symbol_name	= RAS_HOOK_FUNC_SEA,
	.pre_handler	= ras_do_sea_pre,
	.fault_handler	= ras_do_sea_fault,
};

static int register_ras_handler(void)
{
	int ret;

	ret = register_kprobe(&kp_do_serror);
	if (ret) {
		ras_err_msg(ERR_MSG_REGISTER_KPROBE_FAILED,
			    "Failed to hook the function. [%s] (ret=%d)\n",
			    RAS_HOOK_FUNC_SERROR, ret);
		goto out_return_err;
	}

	ret = register_kprobe(&kp_do_sea);
	if (ret) {
		ras_err_msg(ERR_MSG_REGISTER_KPROBE_FAILED,
			    "Failed to hook the function. [%s] (ret=%d)\n",
			    RAS_HOOK_FUNC_SEA, ret);
		goto out_unregister_do_serror;
	}

	return ret;

out_unregister_do_serror:
	unregister_kprobe(&kp_do_serror);

out_return_err:
	return ret;
}

static void unregister_ras_handler(void)
{
	unregister_kprobe(&kp_do_sea);
	unregister_kprobe(&kp_do_serror);
}

/* GIC interrupt handler */
#define GICv3_GICD_BASE_ADDR			(0x002BF00000)
#define GICv3_ITS_CNTL_REG_BASE_ADDR		(0x002BF20000)

#define GICv3_GICD_SIZE				(64 * 1024)
#define GICv3_ITS_CNTL_REG_SIZE			(128 * 1024)

/* register offset */
#define GICD_FJ_GIC_HSWE_PENDING_OFFSET		(0xc000)
#define GITS_FJ_ITS_ERROR_STATUS_OFFSET		(0xc010)
#define GICD_FJ_ICC_ERROR_STATUS_OFFSET		(0xc010)

/* HOST SOFTWARE ERROR */
#define ERROR_PENDING_HSWE_MASK			(0x03)
#define ERROR_PENDING_HSWE(error_pending)	(error_pending & ERROR_PENDING_HSWE_MASK)

#define ERROR_PENDING_NONE			(0x00)
#define ERROR_PENDING_ITS			(0x01)
#define ERROR_PENDING_ICC			(0x02)

static void __iomem *gicd_base = NULL;
static void __iomem *its_cntl_reg_base = NULL;

char *cpuras_dump_gicd = NULL;
char *cpuras_dump_its_cntl = NULL;
atomic_t cpuras_dump_count;

/* host software error handler */
static irqreturn_t host_sw_err_interrupt(int irq, void *dev_id)
{
	u64 error_pending = 0;
	u64 error_status = 0;

	/* read GICD_FJ_GIC_HOST_SOFTWARE_ERROR_PENDING */
	error_pending = readq_relaxed(gicd_base + GICD_FJ_GIC_HSWE_PENDING_OFFSET);

	/* check GICD_FJ_GIC_HOST_SOFTWARE_ERROR_PENDING */
	if (ERROR_PENDING_HSWE(error_pending) & ERROR_PENDING_ITS) {
		/* read GITS_FJ_ITS_ERROR_STATUS */
		error_status = readq_relaxed(its_cntl_reg_base + 
					     GITS_FJ_ITS_ERROR_STATUS_OFFSET);

		if (error_detect_enable & HSWE_DETECT) {
			/* output syslog */
			ras_err_msg(ERR_MSG_ERROR_PENDING_ITS,
				    "ITS HOST SOFTWARE ERROR detected.\n"
				    "GITS_FJ_ITS_ERROR_STATUS=0x%llx\n", error_status);
		}

		/* clear GICD_FJ_GIC_HOST_SOFTWARE_ERROR_PENDING.HSWE.ITS */
		writeq_relaxed(ERROR_PENDING_ITS,
			       gicd_base + GICD_FJ_GIC_HSWE_PENDING_OFFSET);
	}

	if (ERROR_PENDING_HSWE(error_pending) & ERROR_PENDING_ICC) {
		/* read GICD_FJ_ICC_ERROR_STATUS */
		error_status = readq_relaxed(gicd_base + GICD_FJ_ICC_ERROR_STATUS_OFFSET);

		if (error_detect_enable & HSWE_DETECT) {
			/* output syslog */
			ras_err_msg(ERR_MSG_ERROR_PENDING_ICC,
				    "ICC HOST SOFTWARE ERROR detected.\n"
				    "GICD_FJ_ICC_ERROR_STATUS=0x%llx\n", error_status);
		}

		/* clear GICD_FJ_GIC_HOST_SOFTWARE_ERROR_PENDING.HSWE.ICC */
		writeq_relaxed(ERROR_PENDING_ICC,
			       gicd_base + GICD_FJ_GIC_HSWE_PENDING_OFFSET);
	}

	/* check error detect */
	if (!(error_detect_enable & HSWE_DETECT)) {
		/* do not detect error */
		goto out;
	}

	if (ERROR_PENDING_HSWE(error_pending) == ERROR_PENDING_NONE) {
		ras_err_msg(ERR_MSG_HOST_SW_ERR_PENDING_NONE,
			    "A pending error was not detected "
			    "even though an interrupt was received.\n"
			    "GICD_FJ_GIC_HOST_SOFTWARE_ERROR_PENDING=0x%llx\n", error_pending);

		if (hswe_panic_enable == true) {
			panic("host software error pending none.");
		}
	}
	else {
		ras_err_msg(ERR_MSG_HOST_SW_ERR_DETECTED,
			    "HOST SOFTWARE ERROR detected.\n"
			    "GICD_FJ_GIC_HOST_SOFTWARE_ERROR_PENDING=0x%llx\n", error_pending);

		if (hswe_panic_enable == true) {
			panic("host software error detected.\n");
		}
	}

out:
	return IRQ_HANDLED;
}

static void kill_qemu_kvm_process(void)
{
	struct task_struct *task_itr = NULL;

	for_each_process(task_itr) {
		int i = 0;

		ras_debug_msg(DEBUG_MSG_GUEST_SW_ERR_COMM, "comm=%s\n", task_itr->comm);

		/* search guest OS proceess */
		if (!strcmp(task_itr->comm, "qemu-system-aar")) {
			for (i = 0; i < task_itr->files->next_fd; i++) {
				struct file *fd = task_itr->files->fdtab.fd[i];
				const char *dev_str = NULL;

				if (fd == NULL) {
					continue;
				}
				dev_str = fd->f_path.dentry->d_name.name;
				ras_debug_msg(DEBUG_MSG_GUEST_SW_ERR_FDFILE, "fdfile=%s\n", dev_str);

				if (!strcmp(dev_str, "kvm")) {
					/* Forcibly send SIGKILL */
					send_sig(SIGKILL, task_itr, 2);
					break;
				}
			}
		}
	}

	return;
}

/* GUEST SOFTWARE ERROR */
#define ERROR_PENDING_GSWE_MASK			(0x01)
#define ERROR_PENDING_GSWE(error_pending)	(error_pending & ERROR_PENDING_GSWE_MASK)

#define ERROR_PENDING_ICH_ICV			(0x01)

#define GICD_FJ_GIC_GSWE_PENDING_OFFSET		(0xc008)
#define GICD_FJ_ICH_ICV_ERROR_STATUS_OFFSET	(0xc018)

/* guest software error handler */
static irqreturn_t guest_sw_err_interrupt(int irq, void *dev_id)
{
	u64 error_pending = 0;

	/* read GICD_FJ_GIC_GUEST_SOFTWARE_ERROR_PENDING */
	error_pending = readq_relaxed(gicd_base + GICD_FJ_GIC_GSWE_PENDING_OFFSET);

	if (ERROR_PENDING_GSWE(error_pending) == ERROR_PENDING_ICH_ICV) {
		u64 error_status;

		/* read GICD_FJ_ICH_ICV_ERROR_STATUS */
		error_status = readq_relaxed(gicd_base + GICD_FJ_ICH_ICV_ERROR_STATUS_OFFSET);

		/* clear GICD_FJ_GIC_GUEST_SOFTWARE_ERROR_PENDING.GSWE */
		writeq_relaxed(ERROR_PENDING_ICH_ICV,
			       gicd_base + GICD_FJ_ICH_ICV_ERROR_STATUS_OFFSET);

		/* check error detect */
		if (!(error_detect_enable & GSWE_DETECT)) {
			/* do not detect error */
			goto out;
		}

		/* output syslog */
		ras_err_msg(ERR_MSG_ERROR_PENDING_ICH_ICV,
			    "ICH ICV GUEST SOFTWARE ERROR detected.\n"
			    "GICD_FJ_ICH_ICV_ERROR_STATUS=0x%llx\n", error_status);

		/* kill qemu and kvm process */
		kill_qemu_kvm_process();
	}

out:
	return IRQ_HANDLED;
}

/* hwirq to virq */
static int ras_hwirq2virq(const int hwirq, bool *registerd)
{
	int virq = -1, ret = -1;

	ret = acpi_gsi_to_irq(hwirq, &virq);
	if (ret < 0) {
		virq = acpi_register_gsi(NULL, hwirq, ACPI_EDGE_SENSITIVE,
					 ACPI_ACTIVE_HIGH);
		if (virq <= 0) {
			ras_err_msg(ERR_MSG_ACPI_REGISTER_GSI_FAILED,
				    "acpi_register_gsi() failed. ret=%d\n", ret);
		}
		else {
			if (registerd) {
				*registerd = 1;
			}
		}
	}
	return virq;
}

static void cpuras_dump_cpy(uint64_t *dest, const uint64_t *src, size_t size)
{
	int i;
	const int count = size / sizeof(uint64_t);

	for (i = 0; i < count; i++) {
		dest[i] = src[i];
	}
}

int cpuras_dump_handler(struct notifier_block *nb, unsigned long action, void *data)
{
	if (atomic_inc_return(&cpuras_dump_count) == 1) {
		if ((cpuras_dump_gicd != NULL) && (gicd_base != NULL)) {
			cpuras_dump_cpy((uint64_t *)cpuras_dump_gicd, gicd_base, GICv3_GICD_SIZE);
		}

		if ((cpuras_dump_its_cntl != NULL) && (its_cntl_reg_base != NULL)) {
			cpuras_dump_cpy((uint64_t *)cpuras_dump_its_cntl, its_cntl_reg_base,
				GICv3_ITS_CNTL_REG_SIZE);
		}
	}
	return 0;
}

static struct notifier_block cpuras_dump_die_handler = {
	.notifier_call = cpuras_dump_handler,
};

static struct notifier_block cpuras_dump_panic_handler = {
	.notifier_call = cpuras_dump_handler,
};

#define GIC_HOST_SW_ERR_INTR	51
#define GIC_GUEST_SW_ERR_INTR	52
static int host_virq = 0;
static bool host_virq_registered = 0;
static int guest_virq = 0;
static bool guest_virq_registered = 0;

#define READ_ACCESS  (0)
#define WRITE_ACCESS (1)
#define ACCESS_REG_FUNC(name, reg) \
	static void xos_access_##name(uint8_t flag, uint64_t *reg_value){ \
		if(flag == READ_ACCESS){ \
			*reg_value = read_sysreg_s(reg);	\
		} \
		else if(flag == WRITE_ACCESS){ \
			write_sysreg_s(*reg_value, reg); \
		}else{ \
			; \
		}	\
	}

/* Tchipアーキテクチャ判別関数 */
#define XOS_MIDR_EL1		sys_reg(3, 0, 0, 0, 0)
#define XOS_MIDR_EL1_IMPLEMENTER_FJ (0x46)
#define XOS_MIDR_EL1_IMPLEMENTER_MASK (0xFFUL)
#define XOS_MIDR_EL1_IMPLEMENTER_SHIFT (24)
#define XOS_MIDR_EL1_PPNUM_TCHIP (0x1)
#define XOS_MIDR_EL1_PPNUM_MASK (0xFFFUL)
#define XOS_MIDR_EL1_PPNUM_SHIFT (0x4)
ACCESS_REG_FUNC(midr_el1, XOS_MIDR_EL1);

#define XOS_FALSE (0)
#define XOS_TRUE (1)

static int xos_is_tchip(void){

	uint64_t reg = 0;
	int ret=0, impl=0, part=0;

	xos_access_midr_el1(READ_ACCESS, &reg);

	impl = (reg >> XOS_MIDR_EL1_IMPLEMENTER_SHIFT) & XOS_MIDR_EL1_IMPLEMENTER_MASK;
	part = (reg >> XOS_MIDR_EL1_PPNUM_SHIFT) & XOS_MIDR_EL1_PPNUM_MASK;

	if((impl == XOS_MIDR_EL1_IMPLEMENTER_FJ) && (part == XOS_MIDR_EL1_PPNUM_TCHIP)){
		ret = XOS_TRUE;
	}
	else{
		ret = XOS_FALSE;
	}

	return ret;
}

/* initialization */
static int __init cpu_mem_ras_init(void)
{
	int ret = 0;

	if (xos_is_tchip() == XOS_FALSE) {
		return -ENODEV;
	}

	/* moduleparam check */
	if (node_policy == NULL) {
		node_policy = node_policy_str[NODE_POLICY_STR_INDEX_STOP];
		node_policy_flag = NODE_POLICY_STR_INDEX_STOP;
	}
	else {
		const char *cp = NULL;
		struct policy_info node_policy_info;

		node_policy_info.policy_str = (char **)node_policy_str;
		node_policy_info.array_size = ARRAY_SIZE(node_policy_str);
		node_policy_info.policy_flag = &node_policy_flag;
		cp = check_node_policy(node_policy, &node_policy_info);
		if (cp == NULL) {
			return -EINVAL;
		}
	}

	if (uae_store_policy == NULL) {
		uae_store_policy = uae_store_policy_str[UAE_STORE_POLICY_STR_INDEX_LOG];
		uae_store_policy_flag = UAE_STORE_POLICY_STR_INDEX_LOG;
	}
	else {
		const char *cp = NULL;
		struct policy_info uae_store_policy_info;

		uae_store_policy_info.policy_str = (char **)uae_store_policy_str;
		uae_store_policy_info.array_size = ARRAY_SIZE(uae_store_policy_str);
		uae_store_policy_info.policy_flag = &uae_store_policy_flag;
		cp = check_node_policy(uae_store_policy, &uae_store_policy_info);
		if (cp == NULL) {
			uae_store_policy = uae_store_policy_str[UAE_STORE_POLICY_STR_INDEX_LOG];
			uae_store_policy_flag = UAE_STORE_POLICY_STR_INDEX_LOG;
		}
	}

	uae_store_log_wq = create_singlethread_workqueue("uae_store_log");
	if (!uae_store_log_wq) {
		ras_err_msg(ERR_MSG_CREATE_QUEUE_FAILED,
			    "Failed to create workqueue to suppress "
			    "Uncorrectable Address Error (Store) log.\n");
		return -ENOMEM;
	}

	spin_lock_init(&uae_store_log_lock);

	/* register RAS handler setting */
	ret = register_ras_handler();
	if (ret) {
		ras_err_msg(ERR_MSG_REGISTER_RAS_HANDLER,
			    "register_ras_handler failed, "
			    "since the handler has already been registered.\n");
		ret = -EBUSY;
		goto err_register_ras_handler;
	}

	/* ioremap for GICv3_GICD */
	gicd_base = ioremap(GICv3_GICD_BASE_ADDR, GICv3_GICD_SIZE);
	if (gicd_base == NULL) {
		ras_err_msg(ERR_MSG_IOREMAP_FAILD_GICv3_GICD, 
			    "ioremap failed for GICv3_GICD.\n");
		ret = -ENOMEM;
		goto err_ioremap_gicd;
	}

	/* allocate GICv3_GICD dump area */
	cpuras_dump_gicd = kzalloc(GICv3_GICD_SIZE, GFP_ATOMIC);
	if (cpuras_dump_gicd == NULL) {
		ret = -ENOMEM;
		goto err_alloc_dump_gicd;
	}

	/* ioremap for GICv3_ITS_CNTL_REG */
	its_cntl_reg_base = ioremap(GICv3_ITS_CNTL_REG_BASE_ADDR, GICv3_ITS_CNTL_REG_SIZE);
	if (its_cntl_reg_base == NULL) {
		ras_err_msg(ERR_MSG_HOST_SW_ERR_IOREMAP_FAILD_GICv3_ITS_CNTL_REG,
			    "ioremap failed for GICv3_ITS_CNTL_REG.\n"); 
		ret = -ENOMEM;
		goto err_ioremap_its_ctrl_reg;
	}

	/* allocate GICv3_ITS_CNTL_REG dump area */
	cpuras_dump_its_cntl = kzalloc(GICv3_ITS_CNTL_REG_SIZE, GFP_ATOMIC);
	if (cpuras_dump_its_cntl == NULL) {
		ret = -ENOMEM;
		goto err_alloc_dump_its_ctrl;
	}

	atomic_set(&cpuras_dump_count, 0);
	register_die_notifier(&cpuras_dump_die_handler);
	atomic_notifier_chain_register(&panic_notifier_list, &cpuras_dump_panic_handler);

	/* gic host software error handler setting */
	host_virq = ras_hwirq2virq(GIC_HOST_SW_ERR_INTR,
					&host_virq_registered);
	if (host_virq <= 0) {
		ras_err_msg(ERR_MSG_HOST_SW_ERR_VIRQ_FAILED,
			    "Failed to register HOST SOFTWARE ERROR virq.\n");
		ret = -EINVAL;
		goto err_host_virq;
	}
	ras_debug_msg(DEBUG_MSG_GET_HOST_VIRQ,
		      "get host_virq=%d\n", host_virq);

	ret = request_irq(host_virq, host_sw_err_interrupt,
				0, "GIC: HOST_SOFTWARE_ERROR", NULL);
	if (ret) {
		ras_err_msg(ERR_MSG_HOST_SW_ERR_REQUEST_IRQ_FAILED,
			    "HOST SOFTWARE ERROR request_irq failed. (%d)\n", ret);
		goto err_host_request_irq;
	}

	/* gic guest software error handler setting */
	guest_virq = ras_hwirq2virq(GIC_GUEST_SW_ERR_INTR,
					&guest_virq_registered);
	if (guest_virq <= 0) {
		ras_err_msg(ERR_MSG_GUEST_SW_ERR_VIRQ_FAILED,
			    "Failed to register GUEST SOFTWARE ERROR virq.\n");
		ret = -EINVAL;
		goto err_guest_virq;
	}
	ras_debug_msg(DEBUG_MSG_GET_GUEST_VIRQ,
		      "get guest_virq=%d\n", guest_virq);

	ret = request_irq(guest_virq, guest_sw_err_interrupt,
				0, "GIC: GUEST_SOFTWARE_ERROR", NULL);
	if (ret) {
		ras_err_msg(ERR_MSG_GUEST_SW_ERR_REQUEST_IRQ_FAILED,
			    "GUEST SOFTWARE ERROR request_irq failed. (%d)\n", ret);
		goto err_guest_request_irq;
	}

	return 0;

err_guest_request_irq:
	if (guest_virq_registered) {
		acpi_unregister_gsi(GIC_GUEST_SW_ERR_INTR);
	}

err_guest_virq:
	free_irq(host_virq, NULL);

err_host_request_irq:
	if (host_virq_registered) {
		acpi_unregister_gsi(GIC_HOST_SW_ERR_INTR);
	}

err_host_virq:
	atomic_notifier_chain_unregister(&panic_notifier_list, &cpuras_dump_panic_handler);
	unregister_die_notifier(&cpuras_dump_die_handler);
	kfree(cpuras_dump_its_cntl);

err_alloc_dump_its_ctrl:
	iounmap(its_cntl_reg_base);

err_ioremap_its_ctrl_reg:
	kfree(cpuras_dump_gicd);

err_alloc_dump_gicd:
	iounmap(gicd_base);

err_ioremap_gicd:
	unregister_ras_handler();

err_register_ras_handler:
	if (uae_store_log_wq) {
		cancel_delayed_work(&uae_store_log_work);
		flush_workqueue(uae_store_log_wq);
		destroy_workqueue(uae_store_log_wq);
	}

	return ret;
}

/* finalize */
static void __exit cpu_mem_ras_exit(void)
{
	if (uae_store_log_wq) {
		cancel_delayed_work(&uae_store_log_work);
		flush_workqueue(uae_store_log_wq);
		destroy_workqueue(uae_store_log_wq);
	}

	free_irq(guest_virq, NULL);

	if (guest_virq_registered) {
		acpi_unregister_gsi(GIC_GUEST_SW_ERR_INTR);
	}

	free_irq(host_virq, NULL);

	if (host_virq_registered) {
		acpi_unregister_gsi(GIC_HOST_SW_ERR_INTR);
	}

	atomic_notifier_chain_unregister(&panic_notifier_list, &cpuras_dump_panic_handler);
	unregister_die_notifier(&cpuras_dump_die_handler);

	kfree(cpuras_dump_its_cntl);

	/* iounmap for GICv3_ITS_CNTL_REG */
	iounmap(its_cntl_reg_base);

	kfree(cpuras_dump_gicd);

	/* iounmap for GICv3_GICD */
	iounmap(gicd_base);

	unregister_ras_handler();

	return;
}

module_init(cpu_mem_ras_init);
module_exit(cpu_mem_ras_exit);

MODULE_LICENSE("Dual BSD/GPL");
